# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# NOTICE: copied and modified during the summer of 2023
# This code is based on qiskit's own implementation of random_circuit,
# but has been modified.

import numpy as np

from qiskit.circuit import QuantumCircuit, CircuitInstruction
from qiskit.circuit.library import standard_gates


def random_circuit(num_qubits, gates, depth, fillgrade=1):
    if num_qubits == 0:
        return QuantumCircuit()

    possible_gates = {
        # Gate class: (number of qubits, number of parameters)
        standard_gates.IGate: (1, 0),
        standard_gates.SXGate: (1, 0),
        standard_gates.XGate: (1, 0),
        standard_gates.RZGate: (1, 1),
        standard_gates.RGate: (1, 2),
        standard_gates.HGate: (1, 0),
        standard_gates.PhaseGate: (1, 1),
        standard_gates.RXGate: (1, 1),
        standard_gates.RYGate: (1, 1),
        standard_gates.SGate: (1, 0),
        standard_gates.SdgGate: (1, 0),
        standard_gates.SXdgGate: (1, 0),
        standard_gates.TGate: (1, 0),
        standard_gates.TdgGate: (1, 0),
        standard_gates.UGate: (1, 3),
        standard_gates.U1Gate: (1, 1),
        standard_gates.U2Gate: (1, 2),
        standard_gates.U3Gate: (1, 3),
        standard_gates.YGate: (1, 0),
        standard_gates.ZGate: (1, 0),
        standard_gates.CXGate: (2, 0),
        standard_gates.DCXGate: (2, 0),
        standard_gates.CHGate: (2, 0),
        standard_gates.CPhaseGate: (2, 1),
        standard_gates.CRXGate: (2, 1),
        standard_gates.CRYGate: (2, 1),
        standard_gates.CRZGate: (2, 1),
        standard_gates.CSXGate: (2, 0),
        standard_gates.CUGate: (2, 4),
        standard_gates.CU1Gate: (2, 1),
        standard_gates.CU3Gate: (2, 3),
        standard_gates.CYGate: (2, 0),
        standard_gates.CZGate: (2, 0),
        standard_gates.RXXGate: (2, 1),
        standard_gates.RYYGate: (2, 1),
        standard_gates.RZZGate: (2, 1),
        standard_gates.RZXGate: (2, 1),
        standard_gates.XXMinusYYGate: (2, 2),
        standard_gates.XXPlusYYGate: (2, 2),
        standard_gates.ECRGate: (2, 0),
        standard_gates.CSGate: (2, 0),
        standard_gates.CSdgGate: (2, 0),
        standard_gates.SwapGate: (2, 0),
        standard_gates.iSwapGate: (2, 0),
        standard_gates.CCXGate: (3, 0),
        standard_gates.CSwapGate: (3, 0),
        standard_gates.CCZGate: (3, 0),
        standard_gates.RCCXGate: (3, 0),
        standard_gates.C3SXGate: (4, 0),
        standard_gates.RC3XGate: (4, 0),
    }

    gates_1q = []
    extended_gates = []
    for gate in gates:
        n_q, n_p = possible_gates[gate]
        if n_q > num_qubits:
            continue

        extended_gates.append((gate, n_q, n_p))

        if n_q == 1:
            gates_1q.append((gate, n_q, n_p))

    gates = np.array(
        extended_gates,
        dtype=[("class", object), ("num_qubits", np.int64), ("num_params", np.int64)],
    )
    gates_1q = np.array(gates_1q, dtype=gates.dtype)

    qc = QuantumCircuit(num_qubits)

    seed = np.random.randint(0, np.iinfo(np.int32).max)
    rng = np.random.default_rng(seed)

    qubits = np.array(qc.qubits, dtype=object, copy=True)

    filling = np.random.random(size=depth)

    # Apply arbitrary random operations in layers across all qubits.
    for i in range(depth):
        # We generate all the randomness for the layer in one go, to avoid many separate calls to
        # the randomisation routines, which can be fairly slow.

        # This reliably draws too much randomness, but it's less expensive than looping over more
        # calls to the rng. After, trim it down by finding the point when we've used all the qubits.
        gate_specs = rng.choice(gates, size=len(qubits))
        cumulative_qubits = np.cumsum(gate_specs["num_qubits"], dtype=np.int64)
        # Efficiently find the point in the list where the total gates would use as many as
        # possible of, but not more than, the number of qubits in the layer.  If there's slack, fill
        # it with 1q gates.
        max_index = np.searchsorted(cumulative_qubits, num_qubits, side="right")
        gate_specs = gate_specs[:max_index]
        slack = num_qubits - cumulative_qubits[max_index - 1]
        if slack and filling[i] < fillgrade:
            gate_specs = np.hstack((gate_specs, rng.choice(gates_1q, size=slack)))

        # For efficiency in the Python loop, this uses Numpy vectorisation to pre-calculate the
        # indices into the lists of qubits and parameters for every gate, and then suitably
        # randomises those lists.
        q_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
        p_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
        q_indices[0] = p_indices[0] = 0
        np.cumsum(gate_specs["num_qubits"], out=q_indices[1:])
        np.cumsum(gate_specs["num_params"], out=p_indices[1:])
        parameters = rng.uniform(0, 2 * np.pi, size=p_indices[-1])
        rng.shuffle(qubits)

        # We've now generated everything we're going to need.  Now just to add everything.  The
        # conditional check is outside the two loops to make the more common case of no conditionals
        # faster, since in Python we don't have a compiler to do this for us.

        for gate, q_start, q_end, p_start, p_end in zip(
            gate_specs["class"],
            q_indices[:-1],
            q_indices[1:],
            p_indices[:-1],
            p_indices[1:],
        ):
            operation = gate(*parameters[p_start:p_end])
            qc._append(
                CircuitInstruction(operation=operation, qubits=qubits[q_start:q_end])
            )

    return qc



import numpy as np

# from multiprocessing import Pool # could be used for parallel execution of reservoirs in multiplexing approach

from itertools import combinations
from tqdm import tqdm

from quantumreservoirpy.reservoirbase import QReservoir
from quantumreservoirpy.util import memory_to_mean, create_shifted_array
from quantumreservoirpy.statistic import Statistic


class Static(QReservoir):
    def run(self, timeseries, **kwargs):
        transpile = kwargs.pop("transpile", True)
        tqdm_disable = kwargs.pop("tqdm", False)
        self.precision = kwargs.pop("precision", None)
        shots = kwargs.pop("shots", 10**3)

        num_timesteps = len(timeseries)
        self.shots_taken = {}

        for nr in tqdm(
            range(1, self.num_reservoirs + 1),
            desc="Running reservoirs",
            disable=tqdm_disable,
        ):
            circ = self.circuit(
                timeseries,
                merge_registers=False,
                transpile=transpile,
                reservoir_number=nr,
            ).reverse_bits()

            self.stat = {}  # create new statistics
            shots_total = 0
            new_shots = shots

            while True:
                kwargs["shots"] = new_shots
                self._job = self.backend.run(circ, **kwargs)
                counts = self._job.result().get_counts()

                shots_total += new_shots
                # print(shots, shots_total)
                states_list, var_list = self.measurementStatistics(
                    counts, num_timesteps
                )
                # print(shots_total, states_list[0])

                if not self.precision:
                    break
                else:
                    v = np.amax(np.concatenate(var_list))
                    # print("max var=", v)
                    new_shots = int(v / (self.precision) ** 2) - shots_total
                    if new_shots <= 0:
                        break

            states = (
                np.stack(states_list, axis=0)
                if nr == 1
                else np.hstack((states, np.stack(states_list, axis=0)))
            )
            self.variances = (
                np.stack(var_list, axis=0)
                if nr == 1
                else np.hstack((self.variances, np.stack(var_list, axis=0)))
            )
            self.shots_taken[nr] = shots_total

        self.states = states

        return states

    def state(self):
        return self.states

    def predict(self, num_pred, model, from_series, **kwargs):
        kwargs["tqdm"] = True
        timeplex = kwargs.pop("timeplex", 1)
        M = min(num_pred + len(from_series), self.memory)

        predictions = np.zeros(
            num_pred + len(from_series), dtype=np.array(from_series).dtype
        )

        if np.ndim(from_series) == 2:
            prediction_dimension = np.shape(from_series)[1]
            predictions = np.zeros(
                (num_pred + len(from_series), prediction_dimension),
                dtype=np.array(from_series).dtype,
            )

        predictions[: len(from_series)] = from_series

        for i in tqdm(range(num_pred), desc="Predicting..."):
            curidx = len(from_series) + i
            states = self.run(predictions[:curidx][-M:], **kwargs)

            if timeplex > 1:
                states = create_shifted_array(states, timeplex)
            predictions[curidx] = model.predict(states[-1].reshape((1, -1)))

        return predictions  # [-num_pred:]

    def measurementStatistics(self, counts, num_timesteps):
        states_list = []
        var_list = []

        # number of measurements per time step
        num_meas_pt = int(len(list(counts.keys())[0]) / num_timesteps)

        for t in range(num_timesteps):
            sl = []
            vl = []
            indices = range(num_meas_pt)
            for k in range(1, self.degree + 1):
                for O in combinations(indices, k):
                    statkey = str(t) + " " + str(O)
                    self.stat.setdefault(statkey, Statistic())
                    Static.__add_counts(self.stat[statkey], O, counts, t)
                    sl.append(self.stat[statkey].get_E())
                    vl.append(self.stat[statkey].get_Variance())
            states_list.append(np.array(sl))
            var_list.append(np.array(vl))

        return states_list, var_list

    @staticmethod
    def __add_counts(stat, Obs, counts, t):
        # tmp_counts=0
        # tmp_vals=[]
        for key, count in counts.items():
            key_t = key.split()[t]
            val = (
                0.5
                - np.prod(1 - 2.0 * np.array([int(key_t[ind_O]) for ind_O in Obs])) / 2
            )
            stat.add_sample(val, count)
            # tmp_counts+=count
            # tmp_vals.append(val*count)
        # print(" -> ", np.var(np.array(tmp_vals)))


class Incremental(QReservoir):
    def __init__(self, n_qubits, memory=np.inf, backend=None, num_features=-1) -> None:
        super().__init__(n_qubits, memory, backend)

        if num_features > 0:
            self.num_features = num_features

    def run(self, timeseries, **kwargs):
        transpile = kwargs.pop("transpile", True)
        M = min(len(timeseries), self.memory)
        timeseries_splitted = [timeseries[: i + 1][-M:] for i in range(len(timeseries))]

        total = len(timeseries_splitted)

        with tqdm(total=total) as pbar:
            pbar.set_description("Creating circuits...")
            circuits = [
                self.circuit(
                    series, merge_registers=False, transpile=transpile
                ).reverse_bits()
                for series in timeseries_splitted
            ]

            pbar.set_description("Running job...")
            self._job = self.backend.run(circuits, memory=True, **kwargs)

            result = self._job.result()

            if not hasattr(self, "num_features"):
                self.num_features = memory_to_mean(result.get_memory(0)).size

            states = np.zeros((len(timeseries_splitted), self.num_features))

            pbar.set_description("Analyzing... ")
            for idx, _ in enumerate(timeseries_splitted):
                memory = self._job.result().get_memory(idx)

                avg = memory_to_mean(memory)[-self.num_features :]
                states[idx, self.num_features - len(avg) :] = avg
                pbar.update(1)

            return states

    def __run(self, timeseries, **kwargs):
        transpile = kwargs.pop("transpile", True)
        circ = self.circuit(
            timeseries, merge_registers=False, transpile=transpile
        ).reverse_bits()
        self._job = self.backend.run(circ, memory=True, **kwargs)
        results = self._job.result()

        avg = memory_to_mean(results.get_memory())
        if hasattr(self, "num_features"):
            avg = avg[-self.num_features :]
            temp = np.zeros(self.num_features)
            temp[self.num_features - len(avg) :] = avg
            return temp

        return avg

    def predict(self, num_pred, model, from_series, **kwargs):
        M = min(num_pred + len(from_series), self.memory)

        shape = np.array(np.shape(from_series))
        shape[0] += num_pred
        predictions = np.zeros(shape=shape)
        predictions[: len(from_series)] = from_series

        for i in tqdm(range(num_pred), desc="Predicting..."):
            curidx = len(from_series) + i
            states = self.__run(predictions[:curidx][-M:], kwargs=kwargs)
            pred_state = states.reshape((1, -1))
            predictions[curidx] = model.predict(pred_state)

        return predictions[-num_pred:]



import numpy as np

from quantumreservoirpy.util import listify

from qiskit import QuantumCircuit, ClassicalRegister


class InterfaceCircuit(QuantumCircuit):
    def measure_all(self):
        return self.measure(self.qubits)


class ReservoirCircuit(InterfaceCircuit):
    def measure(self, qubit, clbit=None):
        if clbit:
            return super().measure(qubit, clbit)
        else:
            cr = ClassicalRegister(len(listify(qubit)))
            self.add_register(cr)
            return super().measure(qubit, cr)


class CountingCircuit(InterfaceCircuit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0
        self.idx = 0

    def measure(self, qubit):
        num_meas = len(listify(qubit))
        instr = super().measure(qubit, self.count + np.arange(num_meas, dtype=int))
        self.count += num_meas
        return instr



from itertools import combinations
import numpy as np

import inspect

from quantumreservoirpy.util import randomIsing, get_Ising_circuit, limitrange
from quantumreservoirpy.reservoirs import Static


class PartialMeasurement(Static):
    def __init__(
        self,
        n_qubits,
        n_meas,
        memory=np.inf,
        backend=None,
        degree=1,
        basis="X",
        num_reservoirs=1,
        isingparams=None,
        decode=True,# danger zone: this is only for testing
    ) -> None:
        super().__init__(
            n_qubits, memory, backend, degree=degree, num_reservoirs=num_reservoirs
        )
        self.n_meas = n_meas
        self.basis = basis
        self.decode = decode

        if not isingparams:
            steps = 1
            dt = 1.645
            #top = limitrange(list(combinations(range(n_qubits), 2)))
            top = list(combinations(range(n_qubits), 2))
            self.U = {}
            self.isingparams = {}
            for nr in range(1, num_reservoirs + 1):
                (
                    self.U[nr],
                    self.isingparams[nr],
                ) = randomIsing(n_qubits, top, steps, dt)
        else:
            self.U = {}
            for nr in range(1, num_reservoirs + 1):
                self.U[nr] = get_Ising_circuit(n_qubits, isingparams[nr])

    def during(self, circuit, timestep, reservoirnumber):
        # encode
        for k in range(self.n_meas):
            beta = 3**k
            circuit.rx(-beta / 2 * np.pi * timestep, k)
        # circuit.rx(np.pi * timestep, 0)

        # reservoir
        circuit.append(self.U[reservoirnumber], range(self.n_qubits))

        # decode
        if self.basis == "Z":
            circuit.h(range(self.n_meas))
        elif self.basis == "Y":
            circuit.h(range(self.n_meas))
            circuit.sdg(range(self.n_meas))
        circuit.measure(range(self.n_meas))
        if self.decode:
            circuit.reset(range(self.n_meas))



import numpy as np

from sklearn.model_selection import train_test_split

from qiskit import QuantumCircuit, QuantumRegister


def listify(elem):
    try:
        return list(elem)
    except:
        return [elem]


def shift_down(arr, k):
    t, o = arr.shape
    result = np.zeros_like(arr)

    for j in range(o):
        result[k:, j] = arr[: t - k, j]

    return result


def create_shifted_array(a, k):
    t, o = a.shape
    result = np.zeros((t, k * o), dtype=a.dtype)

    for i in range(k):
        result[:, i * o : (i + 1) * o] = shift_down(a, i)

    return result


def memory_to_mean(memory):
    """Utility for analyzing qiskit.Result.get_memory() memory.
    Assumes the data is a list indexed by shots.
    """

    states = [" ".join(mem).split() for mem in memory]
    numb = np.array(states, dtype=int)
    return np.average(numb, axis=0)


def stress_test_models(X_data, y_data, models, test_to_train_ratio=1 / 3, N=100):
    results = np.zeros(len(models))
    for _ in range(N):
        X_train, X_test, y_train, y_test = train_test_split(
            X_data, y_data, test_size=test_to_train_ratio
        )
        for i, model in enumerate(models):
            model.fit(X_train, y_train)
            score = model.score(X_test, y_test)
            results[i] += score
    return results / N


def get_weights(k, a, b):
    return (b - a) * np.random.random_sample(k) + a


class IsingParams:
    def __init__(
        self,
        topology=None,
        trottersteps=None,
        t=None,
        Jx=None,
        Jz=None,
        hx=None,
        hy=None,
        hz=None,
    ):
        self.topology = topology
        self.trottersteps = trottersteps
        self.t = t
        self.Jx = Jx
        self.Jz = Jz
        self.hx = hx
        self.hy = hy
        self.hz = hz


def get_Ising_circuit(n, isingparams):
    q = QuantumRegister(n)
    circuit = QuantumCircuit(q)
    for _ in range(isingparams.trottersteps):
        for i in range(n):
            circuit.rx(isingparams.t * isingparams.hx[i], i)
            circuit.ry(isingparams.t * isingparams.hy[i], i)
            circuit.rz(isingparams.t * isingparams.hz[i], i)
        for i, e in enumerate(isingparams.topology):
            circuit.rzz(isingparams.t * isingparams.Jx[i], e[0], e[1])
            circuit.rxx(isingparams.t * isingparams.Jz[i], e[0], e[1])
    return circuit

def limitrange(topology, dist=2):
    ltop = []
    for tmp in topology:
        if abs(tmp[0]-tmp[1])<=2:
            ltop.append(tmp)
    return ltop


def randomIsing(n, topology, trottersteps, t):
    Jx = get_weights(len(topology), -1, 1)
    Jz = get_weights(len(topology), -1, 1)
    hx = get_weights(n, -0.5, 0.5)
    hy = get_weights(n, -0.5, 0.5)
    hz = get_weights(n, -0.5, 0.5)
    isingparams = IsingParams(topology, trottersteps, t, Jx, Jz, hx, hy, hz)

    circuit = get_Ising_circuit(n, isingparams)

    return circuit, isingparams



# https://stackoverflow.com/questions/33879523/python-how-can-i-generate-a-wav-file-with-beeps
import numpy as np
import wave
import struct

from quantumreservoirpy.util import listify


def midi_to_frequency(note):
    a = 440
    return (a / 32) * (2 ** ((note - 9) / 12))


sample_rate = 44100.0


def silence(duration_milliseconds=500):
    num_samples = duration_milliseconds * (sample_rate / 1000.0)
    return np.zeros(int(num_samples))


def sinewave(freq=440.0, duration_milliseconds=500):
    num_samples = duration_milliseconds * (sample_rate / 1000.0)
    linspc = np.arange(int(num_samples))
    arr = (
        (num_samples - linspc)
        / num_samples
        * np.sin(2 * np.pi * freq * (linspc / sample_rate))
    )

    return arr


def composite(freqs, duration_milliseconds):
    tone = sinewave(freqs[0], duration_milliseconds=duration_milliseconds)
    for freq in freqs[1:]:
        tone += sinewave(freq=freq, duration_milliseconds=duration_milliseconds)
    return tone / max(max(abs(tone)), 1)


def gen_audio(noter, filename="output.wav", BPM=144):
    DUR = 60000 / BPM * 4

    # wav_file=wave.open(filename,"w")

    with wave.open(filename, "w") as wav_file:
        # num = 0
        nchannels = 1
        sampwidth = 2
        # nframes = len(audio)
        comptype = "NONE"
        compname = "not compressed"
        wav_file.setparams((nchannels, sampwidth, sample_rate, 0, comptype, compname))

        for midi, duration in noter:
            dur = DUR * duration
            if midi == "P":
                arr = silence(duration_milliseconds=dur)
            else:
                freqs = [midi_to_frequency(mid) for mid in listify(midi)]
                arr = composite(freqs, duration_milliseconds=dur)
            # num += len(arr)
            for sample in arr:
                wav_file.writeframes(struct.pack("h", int(sample * 32767.0)))
        # wav_file.setnframes(num)



from itertools import combinations, product
import numpy as np
from qiskit import QuantumCircuit, ClassicalRegister, QuantumRegister

from quantumreservoirpy.util import randomIsing, get_Ising_circuit
from quantumreservoirpy.reservoirs import Static


class Stabilizer(Static):
    def __init__(
        self,
        n_qubits,
        n_meas,
        memory=np.inf,
        backend=None,
        degree=1,
        num_reservoirs=1,
        standard=False,
        isingparams=None,
        decode=True,# danger zone: this is only for testing
    ) -> None:
        super().__init__(
            n_qubits + 1, memory, backend, degree=degree, num_reservoirs=num_reservoirs
        )
        self.n_meas = n_meas
        self.decode = decode

        if not isingparams:
            steps = 1
            dt = 1.645
            #top = limitrange(list(combinations(range(n_qubits), 2)))
            top = list(combinations(range(n_qubits), 2))
            self.U = {}
            self.isingparams = {}
            for nr in range(1, num_reservoirs + 1):
                (
                    self.U[nr],
                    self.isingparams[nr],
                ) = randomIsing(n_qubits, top, steps, dt)
        else:
            self.U = {}
            for nr in range(1, num_reservoirs + 1):
                self.U[nr] = get_Ising_circuit(n_qubits, isingparams[nr])

        self.cs = Stabilizer.get_stabilizer_circuits(n_qubits, n_meas, random=False, standard=standard)
        self.decodermap = Stabilizer.build_decoder_map(n_meas + 1, standard=standard)

    def during(self, circuit, timestep, reservoirnumber):
        circuit.barrier()

        # encode
        for k in range(self.n_meas):
            beta = 3**k
            circuit.rx(-beta / 2 * np.pi * timestep, k)
        # circuit.rx(np.pi * timestep, 0)

        # reservoir
        circuit.append(self.U[reservoirnumber], range(self.n_qubits - 1))

        # decode
        cr = ClassicalRegister(self.n_meas)
        circuit.add_register(cr)
        circuit.barrier()
        for j in range(self.n_meas):
            circuit.append(self.cs[j], range(self.n_qubits))
            circuit.measure(circuit.qubits[self.n_qubits - 1], cr[j])
            circuit.barrier()

        if self.decode:
            Stabilizer.apply_operations_for_integers(circuit, cr, self.decodermap)

    @staticmethod
    def binary_array_to_integer(binary_array):
        binary_string = "".join(map(str, binary_array.astype(int)))
        decimal_integer = int(binary_string, 2)
        return decimal_integer

    @staticmethod
    def indices_of_ones(input_list, n):
        indices = [n-1-index for index, value in enumerate(input_list) if value == 1.0]
        return indices # n-1-index because of Endian encoding of qiskit

    @staticmethod
    def get_parity_measurements(x):
        G_list = []
        for i in range(x.shape[0] - 1):
            tmp = np.zeros(x.shape[0])
            tmp[i] = 1
            tmp[i + 1] = 1
            G_list.append(tmp)
        G = np.array(G_list)
        return np.remainder(G @ x, 2)

    @staticmethod
    def build_decoder_map(n, standard):
        decoder = {}
        if standard:
            for origin in list(product((0, 1), repeat=n-1)):
                p = np.array(origin)
                p = Stabilizer.binary_array_to_integer(p)
                flips = Stabilizer.indices_of_ones(origin, n-1)
                decoder[p] = flips
        else:
            for origin in list(product((0, 1), repeat=n)):
                p = Stabilizer.get_parity_measurements(np.array(origin))
                p = Stabilizer.binary_array_to_integer(p)
                flips = Stabilizer.indices_of_ones(origin, n)
                if p in decoder:
                    if len(decoder[p]) > len(flips):
                        decoder[p] = flips
                else:
                    decoder[p] = flips
        return decoder

    @staticmethod
    def get_stabilizer_circuits(n, m, random=True, standard=False):
        circuits = []
        if random:
            rs = random_clifford(n).stabilizer

            for P in sample(rs.to_labels(), m):
                q = QuantumRegister(n + 1)
                circuit = QuantumCircuit(q)
                circuit.reset([n])
                circuit.h(n)
                for i in range(1, len(P)):
                    if P[i] == "X":
                        circuit.cx(q[n], q[i - 1])
                    elif P[i] == "Y":
                        circuit.cy(q[n], q[i - 1])
                    elif P[i] == "Z":
                        circuit.cz(q[n], q[i - 1])
                circuit.h(n)
                circuits.append(circuit)
        elif standard:
            for j in range(m):
                q = QuantumRegister(n + 1)
                circuit = QuantumCircuit(q)
                circuit.reset([n])
                circuit.h(n)
                circuit.cz(q[n], q[j])
                circuit.h(n)
                circuits.append(circuit)
        else:
            for j in range(m):
                q = QuantumRegister(n + 1)
                circuit = QuantumCircuit(q)
                circuit.reset([n])
                circuit.h(n)
                circuit.cz(q[n], q[j])
                circuit.cz(q[n], q[j + 1])
                circuit.h(n)
                circuits.append(circuit)
        return circuits

    @staticmethod
    def apply_operations_for_integers(circ, c, integer_dict):
        with circ.switch(c) as case:
            for key, value in integer_dict.items():
                if value:
                    with case(key):
                        circ.x(value)



import warnings
import numpy as np

from quantumreservoirpy.reservoircircuit import ReservoirCircuit, CountingCircuit

from abc import ABC, abstractmethod

from qiskit_aer import AerSimulator
from qiskit.compiler import transpile as qtranspile
from qiskit.providers import Backend


class BaseReservoir(ABC):
    """
    Interface for custom reservoirs. Please note that one should NOT
    inherit from this class directly, but rather through one of the

    subclasses in reservoir.py.
    """

    def before(self, circuit):
        pass

    # What to do every timestep
    def during(self, circuit, timestep, reservoir_number):
        pass

    # What should happen after the timeseries
    def after(self, circuit):
        pass


class QReservoir(BaseReservoir):
    def __init__(
        self, n_qubits, memory=np.inf, backend=None, degree=1, num_reservoirs=1
    ) -> None:
        try:
            if not issubclass(type(backend), Backend):
                raise TypeError()
            self.backend = backend
        except:
            warnings.warn(
                "Argument 'backend' must inherit from qiskits abstract 'Backend'. Using 'aer_simulator' instead"
            )
            self.backend = AerSimulator()

        self.n_qubits = n_qubits
        self.memory = memory
        self.degree = degree
        self.num_reservoirs = num_reservoirs

    @abstractmethod
    def run(self, timeseries, **kwargs):
        pass

    @abstractmethod
    def predict(self, num_pred, model, from_series, **kwargs):
        pass

    def circuit(
        self, timeseries, merge_registers=False, transpile=True, reservoir_number=1
    ):
        circ = ReservoirCircuit(self.n_qubits)
        circ = self.__build(circ, timeseries, reservoir_number)

        if merge_registers:
            temp_circ = CountingCircuit(self.n_qubits, circ.num_clbits)
            circ = self.__build(temp_circ, timeseries, reservoir_number)
        if transpile:
            circ = qtranspile(circ, self.backend)
        return circ

    @property
    def job(self):
        try:
            return getattr(self, "_job")
        except AttributeError as e:
            raise AttributeError(str(e) + ". Try runnning the reservoir first")

    def __build(self, circ, timeseries, reservoirnumber):
        self.before(circ)
        for t in timeseries:
            self.during(circ, t, reservoirnumber)
        self.after(circ)

        return circ



from setuptools import setup

INSTALL_REQUIRES = [
    "matplotlib",
    "pylatexenc>=2.0",
    "qiskit>=1.0",
    "qiskit-aer>=0.12.0",
    "numpy>=1.21.6",
    "scikit-learn",
    "ipykernel"
]

setup(name='quantumreservoirpy',
      version='0.2',
      packages=['quantumreservoirpy'],
      install_requires=INSTALL_REQUIRES)



import numpy as np
import matplotlib.pyplot as plt


def state_plotter(x, target):
    n_features = len(x[0])
    n_cols = int(np.ceil(np.sqrt(n_features)))
    n_rows = int(np.ceil(n_features / n_cols))

    fig, axes = plt.subplots(
        ncols=n_cols,
        nrows=n_rows,
        figsize=(10 * n_cols, 10 * n_rows),
        sharex="col",
        sharey="row",
    )
    cmap = plt.get_cmap("jet", len(np.unique(target)))

    try:
        axes = axes.flatten()
    except:
        axes = [axes]

    for i, t in enumerate(np.unique(target)):
        mask = t == target
        for idx in range(n_features):
            ax = axes[idx]
            ax.plot(
                np.arange(len(x))[mask],
                x[mask][:, idx],
                marker="o",
                lw=0,
                label=f"{t}",
                c=cmap(i),
            )
            ax.set_title(f"Feature {idx+1}")
            ax.legend(loc="upper right")
    return fig, axes


def feature_plotter(x, target):
    fig, ax = plt.subplots()
    cmap = plt.get_cmap("jet", len(np.unique(target)))
    ax.scatter(x[:, 0], x[:, 1], cmap=cmap, c=target)
    ax.set_xlabel("state 1")
    ax.set_ylabel("state 2")
    return fig, ax
